#!/usr/bin/env python

import os, sys
from os import F_OK, R_OK, W_OK, X_OK
from os.path import exists, join, dirname, isabs
from stat import S_IFDIR
import shutil
import subprocess

from .util import Checker, mount, make_dirs, grab_uid, grab_gid, log_file

USAGE = """

NAME

  rainbow-inject

SYNOPSIS

  rainbow-inject is a security-wrapper for Sugar Activities. Its purpose is to
  launch them in a low-privilege environment so that, if they are exploited,
  they are able to do less harm than would be the case if they ran with the
  full authority of the XO's human operator.

  This isolation is accomplished by running each activity as a separate user
  from a set of users that is known, at system installation time, to have
  tightly restricted access to the file-system.

  The environment that rainbow-inject creates for each activity it starts is
  controlled by an activity-specific security-record created by the
  rainbow-install-bundle. (Not yet implemented.)

ENVIRONMENT VARIABLES

  SUGAR_BUNDLE_PATH  : MUST contain an absolute path to a bundle to launch.

  SUGAR_BUNDLE_ID    : MUST contain the unique bundle-id of the bundle at
                       SUGAR_BUNDLE_PATH.

  RAINBOW_DIR        : Optional; defaults to $HOME/isolation
                       This directory is used by rainbow-inject for all its
                       run-time data: currently, this means uid/gid allocation
                       data and for storing persistent activity-data storage

  RAINBOW_STRACE_LOG : Optional; defaults to Nothing.
                       If set, rainbow-inject will cause strace to write a
                       complete log to stdout.

ARGUMENTS

  Any arguments will be exec'ed verbatim after rainbow-inject properly
  configures its environment.

FILE DESCRIPTORS

  Open file-descriptors will be passed unmodified to the activity being
  launched. However, for the foreseeable future, rainbow-inject _will_ print
  trace information to descriptor 1 and may print error messages to descriptor
  2 for debugging purposes.

RETURN CODES

  Not specified.

"""

def strace(log, log_path, argv, env):
    # cjb thinks we should really be using SystemTap here and refers us to
    # wiki:SystemTap and #83 <MS>
    log('applying strace')
    args = ['/usr/bin/strace', '-f']
    for k, v in env.iteritems():
        args.append('-E')
        args.append('%s=%s' % (k, v))
    args.extend(argv)

    return args, {}

def grab_home(log, spool, bundle_id, owner_uid, owner_gid):
    uid = grab_uid(join(spool, 'uid_pool'))
    gid = grab_gid(join(spool, 'gid_pool'), join(spool, 'bundle_id_to_gid'), bundle_id)
    home = join(spool, 'uid_to_home_dir', str(uid))
    comment = '%s.%d' % (bundle_id, uid)
    assert not exists(home)

    gid_cmd = ['/usr/sbin/groupadd', '-o', '-g', str(gid), str(gid)]
    log('adding group: %s', ' '.join(gid_cmd))
    try:
        subprocess.check_call(gid_cmd)
    except subprocess.CalledProcessError, e:
        if e.returncode is not 9: # we don't care if the group already exists
            raise e

    uid_cmd = ['/usr/sbin/useradd', '-m', '-u', str(uid), '-g', str(gid),
                                    '-c', comment, '-d', home, str(uid)]
    log('adding user: %s', ' '.join(uid_cmd))
    subprocess.check_call(uid_cmd)

    os.chown(home, owner_uid, gid)
    os.chmod(home, 0750)

    return (uid, gid, home)

def configure_home(log, spool, bundle_id, owner_uid, owner_gid, uid, gid, home):
    path = join(spool, 'uid_to_instance_dir', str(uid))
    make_dirs(path, uid, owner_gid, 0770)
    os.chown(path, uid, owner_gid)
    os.chmod(path, 0770)
    os.symlink(path, join(home, 'instance'))

    path = join(spool, 'gid_to_data_dir', str(gid))
    make_dirs(path, owner_uid, gid, 0770)
    os.chown(path, owner_uid, gid)
    os.chmod(path, 0770)
    os.symlink(path, join(home, 'data'))

    make_dirs(join(home, 'tmp'), uid, gid, 0700)
    mount('tmpfs', join(home, 'tmp'), 'tmpfs', 0, "size=1M,nr_inodes=1024")
    # home is 770 <uid>.<owner_gid> and, in the normal implementation, is in a
    # separate namespace.

    xauth = join(home, '.Xauthority')
    shutil.copy('/home/olpc/.Xauthority', xauth)
    os.chown(xauth, uid, gid)

    for frag in ['session.info', 'config', 'owner.key.pub']:
        path = join('.sugar/default/', frag)
        make_dirs(dirname(join(home, path)), uid, owner_gid, 0770)
        shutil.copy(join('/home/olpc', path),
                    join(home, path))
        os.chown(join(home, path), uid, owner_gid)
        os.chmod(join(home, path), 0770)

def launch(log, home, uid, gid, argv, env, cwd):
    env['USER'] = str(uid)
    env['HOME'] = home
    env['XAUTHORITY'] = join(home, '.Xauthority')
    env['SUGAR_ACTIVITY_ROOT'] = home
    env['TMPDIR'] = join(home, 'tmp')

    log('dropping privilege to (%d, %d)', uid, gid)
    os.setgid(gid)
    os.setuid(uid)

    log('chdir to %s' % home)
    os.chdir(cwd)

    log('umask(0)')
    os.umask(0)

    # Since we've dropped privilege, there's nothing to verify about strace_log
    strace_log = env.get('RAINBOW_STRACE_LOG')
    if strace_log:
        argv, env = strace(log, strace_log, argv, env)

    log('about to execve\nargv: %s\nenv: %s', argv, env)
    os.execvpe(argv[0], argv, env)

def check_bundle_id(bundle_id):
    assert bundle_id and all(s.isalnum() for s in bundle_id.split('.'))

def check_bundle_path(uid, gid, path):
    ck = Checker(path, uid, gid)
    assert ck.positive(F_OK | R_OK, 0) and isabs(path)

def check_argv(argv):
    assert argv and len(argv) > 1

def check_cwd(uid, gid, cwd):
    ck = Checker(cwd, uid, gid)
    assert ck.positive(R_OK | X_OK, S_IFDIR)
    assert ck.negative(W_OK, 0)

def check_spool(spool, owner_uid, owner_gid):
    spool_dirs = ('uid_pool', 'gid_pool', 'bundle_id_to_gid',
                 'uid_to_instance_dir', 'gid_to_data_dir', 'uid_to_home_dir')
    for frag in spool_dirs:
        make_dirs(join(spool, frag), owner_uid, owner_gid, 0775)
        ck = Checker(join(spool, frag), owner_uid, owner_gid)
        assert ck.positive(R_OK | W_OK | X_OK, S_IFDIR)

def check_owner(owner_uid, owner_gid):
    assert (owner_uid, owner_gid) == (500, 500)

def check_home_dirs(uid, gid, home):
    for frag in ('instance', 'data', 'tmp'):
        ck = Checker(join(home, frag), uid, gid)
        assert ck.positive(R_OK | W_OK | X_OK, S_IFDIR)

def check_home(uid, gid, home):
    ck = Checker(home, uid, gid)
    assert ck.positive(R_OK | X_OK, S_IFDIR)
    assert ck.negative(W_OK, 0)

def run(log, env, argv, cwd, owner_uid, owner_gid, bundle_path, bundle_id, spool):
    # Note: exceptions are intended to bubble up to the caller and should
    # terminate execution.
    check_bundle_id(bundle_id)
    check_argv(argv)
    check_owner(owner_uid, owner_gid)

    check_spool(spool, owner_uid, owner_gid)

    (uid, gid, home) = grab_home(log, spool, bundle_id, owner_uid, owner_gid)
    configure_home(log, spool, bundle_id, owner_uid, owner_gid, uid, gid, home)

    if cwd is None:
        cwd = home
    check_cwd(uid, gid, cwd)
    check_home_dirs(uid, gid, home)
    check_home_dirs(owner_uid, owner_gid, home)
    check_home(uid, gid, home)
    check_bundle_path(uid, gid, bundle_path)

    launch(log, home, uid, gid, argv, env, cwd)

def main():
    run(log_file(sys.stdout),
        os.environ,
        sys.argv,
        os.environ.get('SUGAR_BUNDLE_PATH'),
        int(os.environ.get('OWNER_UID', 500)), # BUG: Env-var is not authenticated!!!
        int(os.environ.get('OWNER_GID', 500)),
        os.environ.get('SUGAR_BUNDLE_PATH'),
        os.environ.get('SUGAR_BUNDLE_ID'),
        (os.environ.get('RAINBOW_DIR') or '/home/olpc/isolation/1')) #FIXME

if __name__ == '__main__':
    main()
